

% dynamic
:- dynamic(item_pd/1).
:- dynamic(item/2).
:- dynamic(item_class/1).
:- dynamic(pndrule/3).
:- dynamic(pdrule/4).
:- dynamic(arule/3).
:- dynamic(background/3).
:- dynamic(dataset/1).

% multifile
:- multifile(pndrule/3).
:- multifile(pdrule/4).
:- multifile(arule/3).
:- multifile(background/3).

% libraries
:- use_module(library(lists)).

% utility modules
:- use_module(pretty_print).
:- use_module(util).


% move to the working directory
wdir :- working_directory(_, 'dd').


% indirect discrimination inference

indirect(T, M, DB, C, CT_DBC, Beta2Min, UBMin, LBMin, LB, D, B, As) :-
	backgroundconf(Beta2Min, DB, A, _),
	length(DB, L),
	L =< 4,
	ub(T, M, DB, C, CT_DBC, UBMin, _, D, B),
	inference(M, DB, D, B, C, A, CT_DBC, LB, As),
	LB >= LBMin.

inference(B2Min, D, B, C, LB, A)  :-
	background(DB, A1, CT_BDA),
	confidence(CT_BDA, B2),
	B2 >= B2Min,
	osingle(DB, D, B),
	pndrule(DB, C, CT_DBC),
	pndrule(B, C, CT_BC),
	confidence(CT_BC, DELTA),
	confidence(CT_DBC, GAMMA),
	LB is 1/B2*(B2+GAMMA-1)/DELTA,
	merge(A1, D, A).

inference(elift, DB, D, B, C, A, CT_BDC, LB, Aout) :-
	background(DB, A, CT_BDA), 
	pndrule(B, C, CT_BC),
	confidence(CT_BDA, BETA2), 
	confidence(CT_BC, DELTA), 
	confidence(CT_BDC, GAMMA), 
	LB is 1/BETA2*(BETA2+GAMMA-1)/DELTA, 
	merge(A, D, Aout).


inference(slift, DB, D, B, C, A, CT_BDC, LB, Aout) :-
	background(DB, A, CT_BDA), 
	confidence(CT_BDA, BETA2), 
	pndrule(B, C, CT_BC),
	confidence(CT_BDC, GAMMA), 
	LB1 is 1/BETA2*(BETA2+GAMMA-1), 
	CT_BDC = ct(A1, B1),
	CT_BC = ct(X, Y),
	A2 is X - A1,
	B2 is Y - B1,
	M1 is A1 + B1,
	M2 is A2 + B2,
	N2 is M1 + M2 - M1*BETA2,
	BETAP1 is 1 - (1-BETA2)*M1/N2,
	UB2 is 1 - BETAP1 + A2/M2 * BETAP1,
	LB is LB1/UB2,
	merge(A, D, Aout).

ub(T, M, DB, C, CT_DBC, Min, UB, D, B) :-
	pndrule(DB, C, CT_DBC),
	separate(T, DB, D, B),
	pndrule(B, C, CT_BC),
	CT_DBC = ct(A1,B1),
	CT_BC = ct(X,Y),
	A2 is X - A1,
	B2 is Y - B1,
	CT = ct(A1, B1, A2, B2),
	measure(M, CT, UB),
	UB >= Min.

separate(single, DB, D, B) :-
	osingle(DB, D, B).
separate(subset, DB, D, B) :-
	osubset(DB, D, B).

backgroundconf(MinConf, DB, A, CT_DBA) :-
	background(DB, A, CT_DBA),
	confidence(CT_DBA, CN),
	CN >= MinConf.

background(DB, A, CT) :-
	arule(DB, A, CT).

% distribution of PD rules w.r.t. a discrimination measure and pinstances.

test_dist_credit2 :-
	test_distribution_in(slift, class=bad, 2, 0.7),
	test_distribution_in(elift, class=bad, 1.2, 0.7),
	test_distribution_in(slift, class=bad, 2, 0.75),
	test_distribution_in(elift, class=bad, 1.2, 0.75).

test_distribution_in(M, Class, MinSLift, MinP) :-
	distribution( (item(C, Class), 
			check(M, MinSLift, A, B, C, CT), 
			measure(M, CT, SL), 
			pnoinstance(A, B, C, CT, MinP)), SL, 0.05, R),
	P is floor(MinP*100),
	atom_number(Ps, P),
	string_conc(['dist_', M, '_in_', Ps, '.data'], S2),
	dist_print(S2, R).


test_pnoinstance(As, Bs, Cs, CT, MinP) :-
	discrimination(slift, 3, A, B, C, CT), 
	pnoinstance(A, B, C, CT, MinP), 
	itemset_decode(A, As), 
	itemset_decode(B, Bs), 
	item(C, Cs).

test_pinstance(As, Bs, Cs, CT, MinP, Ds, P) :-
	discrimination(slift, 3, A, B, C, CT), 
	pinstance(A, B, C, CT, MinP, D, P), 
	D \= [],
	itemset_decode(A, As), 
	itemset_decode(B, Bs), 
	itemset_decode(D, Ds), 
	item(C, Cs).

% pinstance
	
pinstance(A, B, C, CT_ABC, MinP, D, P) :- 
	coverage(CT_ABC, S_BA),
	confidence(CT_ABC, CN_ABC), 
	arule(BD, A, CT_BDA),
	oremove(BD, B, D),
	support(CT_BDA, S_BDA),
	P1 is S_BDA/S_BA,
	P1 >= MinP,
	pndrule(BD, C, CT_BDC),
	confidence(CT_BDC, CN_BDC), 
	P2 is CN_BDC/CN_ABC,
	P2 >= MinP,
	P is min(P1, P2).


pnoinstance(A, B, C, CT, MinP) :- 
	\+ pinstance(A, B, C, CT, MinP, _, _).


% distribution of PD rules w.r.t. a discrimination measure

test_dist_credit1 :-
	test_distribution_pd(elift, class=bad),
	test_distribution_pd(slift, class=bad).

test_distribution_pd(Measure, Class) :-
	distribution( (item(C, Class), pdrule(_, _, C, CT), measure(Measure, CT, SL)), SL, 0.05, R),
	string_conc(['dist_', Measure, '_all.data'], F),
	dist_print(F, R).


% checking discrimination, affirmative actions and favoritism

discrimination(M, T, A, B, C, CT) :-
	item(C, class=bad),	% bug, this works for credit, should be generalized
	check(M, T, A, B, C, CT).
	
affirmative(M, T, A, B, C, CT) :-
	item(C, class=good),	% bug, this works for credit, should be generalized
	check(M, T, A, B, C, CT).

favoritism(M, T, A, B, C, CT) :-
	affirmative(M, T, A, B, C, CT).

check(M, T, A, B, C, CT) :-
	pdrule(A, B, C, CT),
	measure(M, CT, EL),
	EL >= T.

checkd(M, T, Ades, Bdes, Cdes, CT) :-
	item_class(Cdes),
	item(C, Cdes),
	check(M, T, A, B, C, CT),
	itemset_decode(A, Ades),
	itemset_decode(B, Bdes).

% itemsets, rules, measures

support( ct(A,_,_,_), A ).
support( ct(A,_), A ).

coverage( ct(A,B,_,_), C ) :- C is A + B.
coverage( ct(A,B), C ) :- C is A + B.

confidence( ct(A,B,_,_), CN) :-  
	AB is A + B, 
	AB =\= 0, 
	CN is A/(A+B).
confidence( ct(A,B), CN) :-  
	AB is A + B, 
	AB =\= 0, 
	CN is A/(A+B).

elift( ct(A,B,C,D), EL) :-  
	AC is A + C,
	AC =\= 0, 
	AB is A + B,
	AB =\= 0, 
	N is A+B+C+D,
	EL is (A*N)/(AB*AC).
	
slift( ct(A,B,C,D), SL) :-  
	C =\= 0, 
	AB is A + B, 
	AB =\= 0, 
	CD is C+D, 
	SL is (A*CD)/(AB*C).
	
olift( ct(A,B,C,D), OL) :-  
	C =\= 0,
	B =\= 0, 
	OL is (A*D)/(C*B).
	
measure(elift, CT, V) :-
	elift(CT, V).

measure(slift, CT, V) :-
	slift(CT, V).

measure(olift, CT, V) :-
	olift(CT, V).

rule(B, C, CT) :- 
	pndrule(B, C, CT).
rule(AB, C, ct(X,Y)) :- 
	pdrule(A,B,C,ct(A,B,C,D)),
	merge(A,B,AB),
	X is A+C,
	Y is B+D.

itemset_decode([], []).
itemset_decode([I|Is], [D|Ds]) :- 
	item(I,D), 
	itemset_decode(Is, Ds).

itemset_split([], [], []).
itemset_split([I|Is], [I|A1s], Ds) :-
	item_pd(I), !,
	itemset_split(Is, A1s, Ds).
itemset_split([I|Is], As, [I|D1s]) :-
	itemset_split(Is, As, D1s).


% shorthands

lc10 :-
	test_credit_ruleload(10).
	
lc20 :-
	test_credit_ruleload(20).
	
lcf20 :-
	test_creditf_ruleload(20).
	

% test - german credit arff load & class setting & rules loading

test_credit_ruleload(Support) :-
	test_credit_arffload,
	load_rules(Support).

test_creditf_ruleload(Support) :-
	test_creditf_arffload,
	load_rules(Support).


% extract and load classification and association rules

	
load_rules(Support) :-
	dd_clear_rules,
	atom_number(Sus, Support),
	dataset(File),
	string_conc([File,'/',Sus,'/'], WDir),
	string_concat(WDir,'arule', Arule),
	consult(Arule),
	load_crules(WDir).

load_crules(WDir) :-
	item_class(C),
	item(N, C),
	atom_number(Ns, N),
	string_conc([WDir, 'pdrule_', Ns], PDrules),
	consult(PDrules),
	string_conc([WDir, 'pndrule_', Ns], PNDrules),
	consult(PNDrules),
	fail.

load_crules(_).


extract(Support) :-
	dataset(File),
	atom_number(Sus, Support),
	findall(C, (item_class(A),term_to_atom(A,C)), CList),
	string_conc_sep(CList, ',', Classes),
	findall(C, (item_pd(A),item(A, N),term_to_atom(N,C)), PDList),
	string_conc_sep(PDList, ',', PDItems),
	string_conc(['wDir=',File,'\ndataset=', File, '\nmin_support=', Sus, '\nfp_extract=true\nclasses=', Classes,
		'\npditems=', PDItems, '\nextract_crule=true\nextract_arule=false\nto_prolog=true\nprolog_clear=true'], Props),
	open('dd.props', write, Out),
	write(Out, Props),
	close(Out),
	shell('dd.bat dd.props'),
	load_rules(Support).


% test - german credit arff load & class setting

test_credit_arffload :-
	arff_load('german_credit'),
	set_class([class=good,class=bad]),
	set_pd([personal_status=female_div_or_dep_or_mar,age=gt_52d6]).

test_creditf_arffload :-
	arff_load('german_creditf'),
	set_class([class=good,class=bad]),
	set_pd([personal_status=male_single,age=from_41d4_le_52d6]).

% setting class and PD items


set_class(Is) :-
	retractall(item_class(_)),
	set_class0(Is).

set_class0([]).
set_class0([I|Is]) :-
	assert(item_class(I)),
	set_class0(Is).

set_pd(Is) :-
	retractall(item_pd(_)),
	set_pd0(Is).

set_pd0([]).
set_pd0([I|Is]) :-
	item(N, I),
	assert(item_pd(N)),
	set_pd0(Is).


% arff to prolog

arff_load(File) :-
	string_conc(['ARFF2FIMI.bat ', File, '/', File, '.arff ', File, '/', File, '.fimi ', File, '/', File, '.decode'], Cmd1),
	shell(Cmd1),
	string_conc([File, '/', File, '.decode'], Cmd2),
	dd_clear,
	assert(dataset(File)),
	consult(Cmd2).


% clearing

dd_clear :-
	retractall(dataset(_)),
	retractall(item(_,_)),
	retractall(item_class(_)),
	retractall(item_pd(_)),
	dd_clear_rules.

dd_clear_rules :-
	retractall(pndrule(_,_,_)),
	retractall(pdrule(_,_,_,_)),
	retractall(arule(_,_,_)).

dd_clear_rules(Class) :-
	item(C, Class),
	retractall(pndrule(_,_,C)),
	retractall(pdrule(_,_,C,_)).


